import asyncio
import json
from pathlib import Path
from typing import TYPE_CHECKING, override

from easyrip import log
from easyrip.ripper.param import FONT_SUFFIX_SET
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver

if TYPE_CHECKING:
    from concurrent.futures import Future


class file_watch:
    observer: BaseObserver | None = None

    _main_loop: asyncio.AbstractEventLoop | None = None
    is_in_push_fonts_msg: bool = False
    is_will_push_fonts_msg: bool = False

    @classmethod
    def set_event_loop(cls, loop: asyncio.AbstractEventLoop) -> None:
        """设置主事件循环，需要在异步环境启动后调用"""
        cls._main_loop = loop

    @classmethod
    async def _push_fonts_msg_core(cls) -> None:
        from .font import font_data
        from .web.msg import process_msg, push_msg

        await font_data.refresh_font_data_dict()
        await push_msg(await process_msg(json.dumps({"get_fonts": ""})))

    @classmethod
    async def push_fonts_msg(cls) -> None:
        if cls.is_in_push_fonts_msg:
            cls.is_will_push_fonts_msg = True
            return

        try:
            cls.is_in_push_fonts_msg = True
            await cls._push_fonts_msg_core()
        finally:
            cls.is_in_push_fonts_msg = False

        if cls.is_will_push_fonts_msg:
            cls.is_will_push_fonts_msg = False
            await cls.push_fonts_msg()

    class File_watch_event_handler(FileSystemEventHandler):
        @override
        def on_any_event(self, event: FileSystemEvent) -> None:
            src_path = Path(str(event.src_path))

            if src_path.suffix.lower() not in FONT_SUFFIX_SET:
                return

            log.info(
                f'File watch: {event.event_type}: "{event.src_path}"{f' -> "{event.dest_path}"' if event.dest_path else ""}',
                is_format=False,
            )

            # 线程安全的方式调用异步函数
            if not (file_watch._main_loop and file_watch._main_loop.is_running()):
                log.error(f"file_watch._main_loop = {file_watch._main_loop}")
                return
            future: Future = asyncio.run_coroutine_threadsafe(
                file_watch.push_fonts_msg(), file_watch._main_loop
            )
            # 可选：添加完成回调处理异常
            future.add_done_callback(
                lambda f: f.exception() if not f.cancelled() else None
            )

    file_watch_event_handler = File_watch_event_handler()

    @classmethod
    def stop(cls):
        if cls.observer is not None:
            cls.observer.stop()

    @classmethod
    def start(cls):
        if cls.observer is not None:
            try:
                cls.observer.start()
            except RuntimeError:
                pass

    @classmethod
    async def new_file_watch(cls, *paths: str) -> None:
        cls.stop()

        cls.observer = Observer()
        for path in paths:
            cls.observer.schedule(cls.file_watch_event_handler, path, recursive=True)

        cls.start()

        # 设置当前事件循环
        cls.set_event_loop(asyncio.get_running_loop())



import os
from pathlib import Path

VERSION = "0.3.4"

WIN_FONT_PATHS: tuple[str, ...] = tuple(
    path if Path(path).is_dir() else ""
    for path in (
        os.path.join(os.environ.get("SYSTEMROOT", ""), "Fonts"),
        os.path.join(os.environ.get("LOCALAPPDATA", ""), r"Microsoft\Windows\Fonts"),
    )
)
"""
* [0] = windows font
* [1] = user font
"""



import importlib.resources as pkg_resources
import socket

import uvicorn
from easyrip import log
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from fastapi.staticfiles import StaticFiles

from .msg import process_msg, push_msg_ws_set

app = FastAPI()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()

    # 定时向客户端推送消息
    push_msg_ws_set.add(websocket)

    try:
        while True:
            msg: str = await websocket.receive_text()  # 接收客户端消息
            await websocket.send_text(await process_msg(msg))  # 发送响应给客户端
    except WebSocketDisconnect as e:
        log.info(f"Client disconnected {e.code}: {e.reason}")


app.mount(
    "/",
    StaticFiles(
        directory=str(pkg_resources.files("windows_font_manager") / "static"), html=True
    ),
    name="root",
)


def find_available_port(start_port: int, max_port: int) -> int:
    for port in range(start_port, max_port + 1):
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("0.0.0.0", port))
                return port
        except OSError:
            continue
    raise Exception(
        f"There are no available ports within the range of {start_port}-{max_port}"
    )


async def run_server() -> None:
    host = "0.0.0.0"
    port = find_available_port(8423, 8999)

    config = uvicorn.Config(
        app,
        host=host,
        port=port,
        reload=True,
        log_level="info",
    )

    server = uvicorn.Server(config)
    print(f"Server: {host}:{port}")
    print(f"View: http://127.0.0.1:{port}")

    await server.serve()


async def run() -> None:
    try:
        await run_server()
    except KeyboardInterrupt:
        print("KeyboardInterrupt exit")



import asyncio
import ctypes
from ctypes import wintypes
from pathlib import Path

from easyrip import log
from easyrip.ripper.sub_and_font.font import Font, load_fonts

from .global_val import WIN_FONT_PATHS


def get_windows_error_message(error_code: int) -> str:
    """
    根据Windows错误代码获取错误消息

    Args:
        error_code: Windows错误代码

    Returns:
        str: 错误描述

    """
    try:
        # 使用FormatMessage获取错误描述
        kernel32 = ctypes.WinDLL("kernel32.dll")
        kernel32.FormatMessageW.argtypes = [
            wintypes.DWORD,
            wintypes.LPVOID,
            wintypes.DWORD,
            wintypes.DWORD,
            wintypes.LPWSTR,
            wintypes.DWORD,
            wintypes.LPVOID,
        ]
        kernel32.FormatMessageW.restype = wintypes.DWORD

        FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000
        FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200

        buffer = ctypes.create_unicode_buffer(256)
        kernel32.FormatMessageW(
            FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
            None,
            error_code,
            0,
            buffer,
            ctypes.sizeof(buffer),
            None,
        )

        return buffer.value.strip()
    except Exception:
        return "Can not get detail"


def remove_win_font_resource(font_path: str) -> tuple[bool, str]:
    try:
        # 加载Windows DLL
        gdi32 = ctypes.WinDLL("gdi32.dll")
        user32 = ctypes.WinDLL("user32.dll")

        # 设置函数原型
        gdi32.RemoveFontResourceExW.argtypes = [
            wintypes.LPCWSTR,
            wintypes.DWORD,
            wintypes.LPVOID,
        ]
        gdi32.RemoveFontResourceExW.restype = wintypes.BOOL

        # 调用API移除字体
        result = gdi32.RemoveFontResourceExW(font_path, 0x10, None)

        if result:
            # 发送消息通知系统字体已变更
            user32.SendMessageW(0xFFFF, 0x001D, 0, 0)

            # 建议使用PostMessage进行广播，避免阻塞
            user32.PostMessageW(0xFFFF, 0x001D, 0, 0)

            return True, "Success"
        # 获取详细的错误信息
        error_code = ctypes.GetLastError()
        error_msg = get_windows_error_message(error_code)

    except Exception as e:
        return False, f"{remove_win_font_resource.__name__} error: {e}"

    return False, f"{remove_win_font_resource.__name__} faild {error_code}: {error_msg}"


def remove_font_reg(path: Path) -> tuple[bool, str]:
    import winreg

    hkey: int | None = None

    if path.parent.samefile(WIN_FONT_PATHS[0]):
        hkey = winreg.HKEY_LOCAL_MACHINE
    elif path.parent.samefile(WIN_FONT_PATHS[1]):
        hkey = winreg.HKEY_CURRENT_USER

    if hkey is None:
        return True, "Not a installation directory"

    try:
        key = winreg.OpenKey(
            hkey,
            r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts",
            access=winreg.KEY_WRITE | winreg.KEY_READ,
        )
    except OSError as e:
        return False, f"faild to openkey: {e}"

    delete_keys: list[str] = []
    try:
        i = 0
        while True:
            name, value, _type = winreg.EnumValue(key, i)
            if path.name.lower() == Path(value).name.lower():
                delete_keys.append(name)
            i += 1
    except OSError:
        pass

    for name in delete_keys:
        try:
            winreg.DeleteValue(key, name)
        except OSError as e:
            return False, f"Failed to delete registry key {name}: {e}"

    return True, f"Successfully removed {len(delete_keys)} registry entries"


def remove_font(path: str | Path) -> tuple[bool, str]:
    path = Path(path)
    if not path.is_file():
        return False, "not a file"

    # 卸载资源
    remove_win_font_resource_res = remove_win_font_resource(str(path))
    if not remove_win_font_resource_res[0]:
        log.warning(remove_win_font_resource_res[1])

    # 删除文件
    try:
        path.unlink()
    except Exception as e:
        return False, f"paht.unlink faild: {e}"

    # 删除注册表
    try:
        remove_font_reg_res = remove_font_reg(path)
    except Exception as e:
        return False, f"{remove_font_reg.__name__} error: {e}"
    if not remove_font_reg_res[0]:
        return False, f"{remove_font_reg.__name__} faild: {remove_font_reg_res[1]}"

    return True, remove_font_reg_res[1]


class font_data:
    font_data_dict: dict[str, list[Font]] = {}
    """实时变化的监听目录及其结果"""

    @classmethod
    async def refresh_font_data_dict(cls):
        log.info(cls.refresh_font_data_dict.__name__)
        cls.font_data_dict = {
            d: await asyncio.to_thread(load_fonts, d) for d in cls.font_data_dict
        }

    @classmethod
    async def add_font_data_dir(cls, *dirs: str):
        from .file_watch import file_watch

        for d in dirs:
            cls.font_data_dict[d] = load_fonts(d)

        await file_watch.new_file_watch(*dirs)

    @classmethod
    def pop_font_data_dir(cls, *dirs: str):
        for d in dirs:
            try:
                cls.font_data_dict.pop(d)
            except KeyError as e:
                log.error(
                    "{} faild: {}",
                    cls.pop_font_data_dir.__name__,
                    e,
                )

    @classmethod
    def get_font_info(
        cls,
        pathname: str,
    ):  # TODO
        fonts = load_fonts(pathname)



import asyncio
import itertools
import json
import weakref
from functools import cache
from pathlib import Path
from typing import Any, Final, TypedDict

from easyrip import log
from easyrip.easyrip_web.third_party_api import github
from easyrip.utils import type_match
from fastapi import WebSocket

from ..file_watch import file_watch
from ..font import font_data, remove_font
from ..global_val import VERSION

push_msg_ws_set: Final[weakref.WeakSet[WebSocket]] = weakref.WeakSet()
"""需要主动推送的 ws"""


async def push_msg(msg: str):
    """主动推送消息"""
    for ws in push_msg_ws_set:
        await ws.send_text(msg)


class Response_font_dict(TypedDict):
    pathname: str
    filename: str
    familys: list[str]
    font_type: str
    font_type_val: tuple[bool, bool]


class Response_font_info_dict(Response_font_dict):
    pass


class Response_info(TypedDict):
    version: str
    latest_release_ver: str | None


class Response_dict(TypedDict):
    """None 表示不返回而不是返回为空"""

    info: Response_info
    execute: list[str]
    fonts: dict[str, list[Response_font_dict]] | None
    fonts_info: Response_font_info_dict | None


class Webui_msg_dict(TypedDict):
    wait_ms: int
    get_fonts: str
    del_fonts: list[str]
    add_dirs: list[str]
    pop_dirs: list[str]
    get_font_info: Any  # TODO


@cache
def _get_github_latest_release_ver(release_api_url: str) -> str | None:
    return github.get_latest_release_ver(release_api_url)


async def get_github_latest_release_ver(release_api_url: str) -> str | None:
    return _get_github_latest_release_ver(release_api_url)


async def process_msg(msg: str) -> str:
    """接收前端消息，返回后端消息"""
    res: Response_dict = {
        "info": {
            "version": VERSION,
            "latest_release_ver": await get_github_latest_release_ver(
                "https://api.github.com/repos/op200/windows_font_manager/releases/latest"
            ),
        },
        "execute": [],
        "fonts": None,
        "fonts_info": None,
    }

    data: Webui_msg_dict = json.loads(msg)

    def write_fonts() -> None:
        res["fonts"] = {
            dir_str: [
                {
                    "pathname": font.pathname,
                    "filename": Path(font.pathname).name,
                    "familys": list(font.familys),
                    "font_type": font.font_type.name,
                    "font_type_val": font.font_type.value,
                }
                for font in font_list
            ]
            for dir_str, font_list in font_data.font_data_dict.items()
        }

    for key, val in data.items():
        if key not in Webui_msg_dict.__annotations__:
            log.error("Unknown key: {}", key)
            continue
        res["execute"].append(key)
        match key:
            case "wait_ms":
                assert isinstance(val, int)

                if val >= 0:
                    log.info("Wait {}ms", val)
                    await asyncio.sleep(val / 1000)
                else:
                    log.warning("The '{}: {}' is illegal", key, val)

            case "get_fonts":
                write_fonts()

            case "del_fonts":
                assert type_match(val, list[str])
                # 解除占用
                for font in itertools.chain.from_iterable(
                    font_data.font_data_dict.values()
                ):
                    if any(Path(del_font).samefile(font.pathname) for del_font in val):
                        # if any(map(Path(font.pathname).samefile, val)):
                        font.font.close()
                        continue
                file_watch.stop()
                # 删除文件
                for pathname in val:
                    log.info("Delete font: {}", pathname)
                    try:
                        remove_res = remove_font(pathname)
                        if not remove_res[0]:
                            raise Exception(remove_res[1])
                    except Exception as e:
                        log.error("Delete font faild: {}", e)
                    else:
                        log.info("Delete font success: {}", remove_res[1])
                # 刷新
                await font_data.refresh_font_data_dict()
                file_watch.start()
                write_fonts()

            case "add_dirs":
                assert type_match(val, list[str])
                await font_data.add_font_data_dir(*val)
                write_fonts()

            case "pop_dirs":
                assert type_match(val, list[str])
                font_data.pop_font_data_dir(*val)
                write_fonts()

            case "get_font_info":
                pass
                # TODO
                # res["fonts_info"] = {}

    return json.dumps(res)



import asyncio

from easyrip import log

from .font import font_data
from .global_val import WIN_FONT_PATHS
from .web import server


async def main():
    log.write_level = log.LogLevel.none
    log.init()

    await font_data.add_font_data_dir(*WIN_FONT_PATHS)
    await server.run()


def run():
    asyncio.run(main())


if __name__ == "__main__":
    run()



import os
from pathlib import Path

from setuptools.command.build import build

DIR = "windows-font-manager-webui"
PNPM_HEAD = f"pnpm --dir ./{DIR}"


def run_pnpm() -> None:
    for param in (
        "i --frozen-lockfile",
        "build",
    ):
        cmd = f"{PNPM_HEAD} {param}"
        print(f"run: {cmd}")
        if os.system(cmd):
            raise RuntimeError(f"{cmd} error")


class Build(build):
    def run(self) -> None:
        if not Path(DIR).is_dir():
            print(f'The dir "{DIR}" not found, do not run pnpm')
            return

        run_pnpm()

        super().run()
